package org.databinding;

import java.beans.PropertyChangeListener;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Set;

import org.databinding.listener.ModelListener;
import org.databinding.util.DBConstants;
import org.databinding.util.PrimitiveWrapperConverter;
import org.databinding.websocket.Message;
import org.databinding.websocket.MessageHandler;
import org.databinding.websocket.Servlet;
import org.databinding.websocket.Socket;
import org.eclipse.jetty.util.ajax.JSON;
import org.eclipse.jetty.util.ajax.JSONObjectConvertor;

public class DataBindingContext {

	private HashMap<String, Binding> bindings = new HashMap<String, Binding>();
	private Servlet servlet;
	private Socket sendingSocket;
	private JSON json;

	public DataBindingContext(Servlet servlet) {
		this.servlet = servlet;
		json = new JSON();
		json.addConvertor(Message.class, new JSONObjectConvertor());
	}

	public void addBinding(Binding binding) {
		this.bindings.put(binding.getId(), binding);

		// reflectively retrieve method to add property listener
		Class<?> parameterTypes[] = new Class[2];
		parameterTypes[0] = String.class;
		parameterTypes[1] = PropertyChangeListener.class;
		Method addListenerMethod = null;
		try {
			addListenerMethod = binding.getModel().getClass()
					.getMethod("addPropertyChangeListener", parameterTypes);
		} catch (Exception e) {
			e.printStackTrace();
		}

		// invoke addPropertyChangeListener(String fieldName,
		// PropertyChangeListener listener)
		try {
			addListenerMethod.invoke(binding.getModel(), new Object[] {
					binding.getField().getName(),
					new ModelListener(this, binding) });
		} catch (Exception e) {
			e.printStackTrace();
		}

	}

	public Binding getBinding(String id) {
		return this.bindings.get(id);
	}

	public HashMap<String, Binding> getBindings() {
		return this.bindings;
	}

	public void updateModel(MessageHandler msgHander) {

		Binding binding = this.bindings.get(msgHander.getMsg().getId());

		if (binding == null) {
			System.err.println("There is no binding associated with the ID "
					+ msgHander.getMsg().getId() + ".");
			return;
		}

		this.sendingSocket = msgHander.getSendingSocket();
		String methodName = getSetterName(binding);
		Class<?> fieldClass = binding.getField().getType();
		Class<?> modelClass = binding.getModel().getClass();

		// retrieve wrapper class
		if (fieldClass.isPrimitive()) {
			fieldClass = PrimitiveWrapperConverter.getWrapper(fieldClass);
		}

		Method setter = getSetterMethod(modelClass, methodName, fieldClass);
		invokeSetterMethod(fieldClass, setter, binding.getModel(), msgHander
				.getMsg().getValue());
	}

	public void updateUI(Object model, Field field) {

		List<Binding> relevantBindings = getBindings(model, field);

		for (Binding binding : relevantBindings) {
			Object value = invokeGetterMethod(getGetterMethod(binding), model);
			if (value != null) {

				Set<Socket> sockets = servlet.getMembers();
				for (Socket socket : sockets) {
					if (sendingSocket != null && sendingSocket.equals(socket)) {
						sendingSocket = null;
					} else {
						Message msg = new Message(binding.getId(),
								value.toString());
						socket.sendMsg(json.toJSON(msg));
//						System.out.println("sent JSON String: "
//								+ json.toJSON(msg));
					}
				}

			}
		}
	}

	public void initializeUI() {

		for (String key : this.bindings.keySet()) {
			Binding binding = bindings.get(key);
			Object value = invokeGetterMethod(getGetterMethod(binding),
					binding.getModel());
			if (value != null) {
				Set<Socket> sockets = servlet.getMembers();
				for (Socket socket : sockets) {
					Message msg = new Message(binding.getId(), value.toString());
					socket.sendMsg(json.toJSON(msg));
//					System.out.println("sent JSON String: " + json.toJSON(msg));
				}
			}
		}
	}

	private String getSetterName(Binding binding) {
		String fieldName = binding.getField().getName();
		String methodName = DBConstants.SETTER_PREFIX
				+ fieldName.substring(0, 1).toUpperCase();
		methodName += fieldName.substring(1);
		return methodName;
	}

	private String getGetterName(Binding binding, String prefix) {
		String fieldName = binding.getField().getName();
		String methodName = prefix + fieldName.substring(0, 1).toUpperCase();
		methodName += fieldName.substring(1);
		return methodName;
	}

	private void invokeSetterMethod(Class<?> fieldClass, Method setter,
			Object model, String value) {
		try {
			if (value != null) {

				// get the wrapper class
				Class wrapperFieldClass = fieldClass;
				if (fieldClass.isPrimitive()) {
					wrapperFieldClass = PrimitiveWrapperConverter
							.getWrapper(fieldClass);
				}

				// get the default String constructor
				Constructor<?> constructor = wrapperFieldClass
						.getConstructor(new Class[] { String.class });

				// call constructor
				Object newValue = constructor.newInstance(value);

				// if the method param is a primitive, the wrapper has to be
				// converted to the primitive representation
				if (setter.getParameterTypes()[0].isPrimitive()) {

					Class primitive = PrimitiveWrapperConverter
							.getPrimitive(newValue.getClass());
					// compose the convert-to-primitive method (e.g. intValue,
					// booleanValue)
					String methodName = primitive.getName() + "Value";
					Method method = newValue.getClass().getMethod(methodName);
					// invoke the convert-to-primitive method
					setter.invoke(model, method.invoke(newValue));
				} else {
					// try to invoke method with wrapper class
					setter.invoke(model, newValue);
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private Object invokeGetterMethod(Method getter, Object model) {
		try {
			return getter.invoke(model);
		} catch (Exception e) {
			e.printStackTrace();
		}

		return null;
	}

	private Method getSetterMethod(Class<?> modelClass, String methodName,
			Class<?> fieldClass) {

		Method setter = null;
		try {
			// try to find method with Wrapper class parameter
			setter = modelClass.getMethod(methodName,
					new Class[] { fieldClass });
		} catch (NoSuchMethodException e) {
			// no exception handling required
		} catch (SecurityException e) {
			e.printStackTrace();
		}

		try {
			if (setter == null) {
				// try to find method with primitive parameter
				Class[] args = new Class[] { PrimitiveWrapperConverter
						.getPrimitive(fieldClass) };
				setter = modelClass.getMethod(methodName, args);
			}

		} catch (NoSuchMethodException e) {
			// no exception handling required
		} catch (SecurityException e) {
			e.printStackTrace();
		}

		return setter;
	}

	private Method getGetterMethod(Binding binding) {
		String methodName = getGetterName(binding, DBConstants.GETTER_PREFIX);
		Method method = null;
		try {
			method = binding.getModel().getClass().getMethod(methodName);
		} catch (SecurityException e) {
			e.printStackTrace();
		} catch (NoSuchMethodException e) {
			/**
			 * exception may be reached if the method is not named after scheme
			 * get+property e.g. boolean properties are typically name according
			 * to the convention is+property
			 */
		}

		if (method != null)
			return method;

		if (binding.getField().getType().equals(boolean.class)
				|| binding.getField().getType().equals(Boolean.class)) {
			methodName = getGetterName(binding, DBConstants.IS_PREFIX);
			try {
				method = binding.getModel().getClass().getMethod(methodName);
			} catch (SecurityException e) {
				e.printStackTrace();
			} catch (NoSuchMethodException e) {
				e.printStackTrace();
			}
		}

		return method;
	}

	private List<Binding> getBindings(Object model, Field field) {

		ArrayList<Binding> relevantBindings = new ArrayList<Binding>();

		for (String key : bindings.keySet()) {
			Binding binding = bindings.get(key);
			if (binding.getModel().equals(model)
					&& binding.getField().equals(field)) {
				relevantBindings.add(binding);
			}
		}

		return relevantBindings;
	}
}